from lclpy.localsearch.simulatedannealing.abstract_cooling_function \
    import AbstractCoolingFunction


class GeometricCoolingFunction(AbstractCoolingFunction):
    """Implements a geometric cooling function for simulated annealing.

    Parameters
    ----------
    alpha : float, optional
        A parameter that influences the cooling rate. Should be a positive
        number smaller than 1. Default is 0.75.

    Attributes
    ----------
    _alpha : float
        A parameter that influences the cooling rate.

    Examples
    --------
    A simple example:

    .. doctest::

        >>> from lclpy.localsearch.simulatedannealing.geometric_cooling_function \\
        ...     import GeometricCoolingFunction
        ... # init
        >>> test = GeometricCoolingFunction()
        ... # tests
        >>> test.next_temperature(1000)
        750.0
        >>> test.next_temperature(100)
        75.0
        >>> test.next_temperature(10)
        7.5

    An example with an alpha of 0.5:

    .. doctest::

        >>> from lclpy.localsearch.simulatedannealing.geometric_cooling_function \\
        ...     import GeometricCoolingFunction
        ... # init
        >>> test = GeometricCoolingFunction(0.5)
        ... # tests
        >>> test.next_temperature(1000)
        500.0
        >>> test.next_temperature(100)
        50.0
        >>> test.next_temperature(10)
        5.0

    """

    def __init__(self, alpha=0.75):
        super().__init__()
        self._alpha = alpha

    def next_temperature(self, old_temperature):
        """Function to get the next temperature.

        Parameters
        ----------
        old_temperature : int or float
            The old temperature.

        Returns
        -------
        float
            The new temperature. This temperature is equal to:

            .. math::
                \\_alpha*old\\_temperature

            .. The formula:
                _alpha*old_temperature

        """

        return self._alpha * old_temperature



"""The package benchmark is designed to benchmark.

The package contains:

- The function benchmark in the module benchmark.
  This function can be used to run multiple algorithms on multiple problems for
  multiple runs

- The module statistics. This module contains several functions to get
  statistical data from the results of the function benchmark.

"""



"""This module contains functions to check if a value is an improvement."""


def bigger(old_value, value):
    """Checks if a value is an improvement. Bigger values are improvements.

    Parameters
    ----------
    old_value : int or float
        The old value.
    value : int or float
        The new value that is checked.

    Returns
    -------
    bool
        Returns true if the new value is bigger than the old value, else
        returns false.

    Examples
    --------
    Some simple tests:

    .. doctest::

        >>> from lclpy.aidfunc.is_improvement_func import bigger
        >>> bigger(2, 5)
        True
        >>> bigger(5, 2)
        False
        >>> bigger(3,3)
        False

    """

    return value > old_value


def smaller(old_value, value):
    """Checks if a value is an improvement. Smaller values are improvements.

    Parameters
    ----------
    old_value : int or float
        The old value.
    value : int or float
        The new value that is checked.

    Returns
    -------
    bool
        Returns true if the new value is smaller than the old value, else
        returns false.

    Examples
    --------
    Some simple tests:

    .. doctest::

        >>> from lclpy.aidfunc.is_improvement_func import smaller
        >>> smaller(2, 5)
        False
        >>> smaller(5, 2)
        True
        >>> smaller(3,3)
        False

    """

    return value < old_value


def bigger_or_equal(old_value, value):
    """Checks if a value is an improvement. Bigger values are improvements.

    Parameters
    ----------
    old_value : int or float
        The old value.
    value : int or float
        The new value that is checked.

    Returns
    -------
    bool
        Returns true if the new value is bigger or equal to the old value, else
        returns false.

    Examples
    --------
    Some simple tests:

    .. doctest::

        >>> from lclpy.aidfunc.is_improvement_func import bigger_or_equal
        >>> bigger_or_equal(2, 5)
        True
        >>> bigger_or_equal(5, 2)
        False
        >>> bigger_or_equal(3,3)
        True

    """

    return value >= old_value


def smaller_or_equal(old_value, value):
    """Checks if a value is an improvement. Smaller values are improvements.

    Parameters
    ----------
    old_value : int or float
        The old value.
    value : int or float
        The new value that is checked.

    Returns
    -------
    bool
        Returns true if the new value is smaller or equal to the old value,
        else returns false.

    Examples
    --------
    Some simple tests:

    .. doctest::

        >>> from lclpy.aidfunc.is_improvement_func import smaller_or_equal
        >>> smaller_or_equal(2, 5)
        False
        >>> smaller_or_equal(5, 2)
        True
        >>> smaller_or_equal(3,3)
        True

    """

    return value <= old_value



from abc import ABC, abstractmethod


class AbstractMove(ABC):
    """Template to create Move-objects.

    This object is used to explore a neighbourhood and alter the state of a
    solution. It can also be used to generate valid moves in the neighbourhood.

    """

    def __init__(self):
        super(AbstractMove, self).__init__()

    def get_move_type(self):
        """Returns the move type.

        This function is used in the implementation of delta evaluation. If
        this is not implemented or wanted, this method does not need to be
        implemented.

        Returns
        -------
        str
            The move type.

        """

        raise NotImplementedError

    @abstractmethod
    def move(self, data, move):
        """Performs a move.

        Parameters
        ----------
        data
            The dataset that is being explored. It will be altered after the
            method call.
        move : tuple of int
            A representation of a valid move.

        """

        pass

    @abstractmethod
    def undo_move(self, data, move):
        """Undoes a move.

        Make sure you only undo a move after it's has been performed.

        Parameters
        ----------
        data
            The dataset that is being explored. It will be altered after the
            method call.
        move : tuple of int
            A representation of the move one wishes to undo

        """

        pass

    @abstractmethod
    def get_moves(self):
        """A generator used to return all valid moves in the neighbourhood.

        Yields
        ------
        tuple of int
            The next valid move.

        """

        pass

    @abstractmethod
    def get_random_move(self):
        """A method used to generate a random move in the neighbourhood.

        Returns
        -------
        tuple of int
            A random valid move.
        """

        pass



from lclpy.termination.abstract_termination_criterion \
    import AbstractTerminationCriterion


class MinTemperatureTerminationCriterion(AbstractTerminationCriterion):
    """A termination criterion for use with a simulated annealing algorithm.

    This termination criterion ends the algorithm if the current temperature
    is lower than the minimal temperature.

    Parameters
    ----------
    min_temperature : int, optional
        The minimal temperature. The default is 10.

    Attributes
    ----------
    min_temperature : int
        The minimal temperature.
    _run : bool
        True if no temperature lower than the minimal has been encountered,
        False if a temperature lower than the minimal has been encountered.

    Examples
    --------
    A simple example of the default behaviour:

    .. doctest::

        >>> from lclpy.termination.min_temperature_termination_criterion \\
        ...     import MinTemperatureTerminationCriterion
        ... # init
        >>> test = MinTemperatureTerminationCriterion()
        ... # tests
        >>> test.keep_running()
        True
        >>> test.check_variable(1000.0)
        >>> test.keep_running()
        True
        >>> test.check_variable(10.0)
        >>> test.keep_running()
        True
        >>> test.check_variable(9.0)
        >>> test.keep_running()
        False
        >>> test.keep_running()
        False

    Example with a minimal temperature of 100:

    .. doctest::

        >>> from lclpy.termination.min_temperature_termination_criterion \\
        ...     import MinTemperatureTerminationCriterion
        ... # init
        >>> test = MinTemperatureTerminationCriterion(100)
        ... # tests
        >>> test.keep_running()
        True
        >>> test.check_variable(1000.0)
        >>> test.keep_running()
        True
        >>> test.check_variable(100.0)
        >>> test.keep_running()
        True
        >>> test.check_variable(10.0)
        >>> test.keep_running()
        False

    """

    def __init__(self, min_temperature=10):
        super().__init__()
        self._min_temperature = min_temperature
        self._run = True

    def keep_running(self):
        """function to determine if the algorithm needs to continue running.

        Returns
        -------
        bool
            The function returns true if the current temperature is higher than
            or equal to the minimal temperature, else it returns false.

        """

        return self._run

    def check_variable(self, current_temperature):
        """Checks if the current temperature is lower than the minimal temperature.

        This function should always be called if the temperature is lowered.

        Parameters
        ----------
        current_temperature : int or float
            The current temperature of the algorithm.

        """

        if self._run:
            self._run = current_temperature >= self._min_temperature

    def reset(self):
        """Resets the object back to it's state after init.

        Examples:

        A simple example with the default min_temperature:

        .. doctest::

            >>> from lclpy.termination.min_temperature_termination_criterion \\
            ...     import MinTemperatureTerminationCriterion
            ... # init
            >>> test = MinTemperatureTerminationCriterion()
            ... # run 1
            >>> test.keep_running()
            True
            >>> test.check_variable(1000.0)
            >>> test.keep_running()
            True
            >>> test.check_variable(10.0)
            >>> test.keep_running()
            True
            >>> test.check_variable(9.0)
            >>> test.keep_running()
            False
            >>> test.keep_running()
            False
            >>> # reset
            >>> test.reset()
            ... # run 2
            >>> test.keep_running()
            True
            >>> test.check_variable(1000.0)
            >>> test.keep_running()
            True
            >>> test.check_variable(10.0)
            >>> test.keep_running()
            True
            >>> test.check_variable(9.0)
            >>> test.keep_running()
            False
            >>> test.keep_running()
            False


        """

        self._run = True




def pass_func(*args):
    """A function that does nothing.

    Parameters
    ----------
    args
        Variable length argument list, it does not matter what or how many
        parameters you use. Does not accept keyword arguments.

    """

    pass



from lclpy.termination.abstract_termination_criterion \
    import AbstractTerminationCriterion


class MultiCriterion(AbstractTerminationCriterion):
    """Class to combine multiple terminationcriteria.

    Parameters
    ----------
    criteria : list or tuple of AbstractTerminationCriterion
        An iterable object containing the intialised termination criterions one
        wishes to use.

    Attributes
    ----------
    criteria : list or tuple of AbstractTerminationCriterion
        An iterable object containing the intialised termination criterions one
        wishes to use.

    Examples
    --------
    3 termination criteria are used, three tests are done to ensure that all
    three criterions are capable of stopping the iterating correctly.

    MaxSecondsTerminationCriterion stops the iterating:

     .. doctest::

        >>> import time
        >>> from lclpy.termination.max_seconds_termination_criterion \\
        ...     import MaxSecondsTerminationCriterion
        >>> from lclpy.termination.max_iterations_termination_criterion \\
        ...     import MaxIterationsTerminationCriterion
        >>> from lclpy.termination.no_improvement_termination_criterion \\
        ...     import NoImprovementTerminationCriterion
        >>> from lclpy.termination.multi_criterion import MultiCriterion
        ... # init list
        >>> criteria = []
        >>> criteria.append(MaxSecondsTerminationCriterion(3))
        >>> criteria.append(MaxIterationsTerminationCriterion(10))
        >>> criteria.append(NoImprovementTerminationCriterion(3))
        ... # init MultiCriterion
        >>> multi_criterion = MultiCriterion(criteria)
        ... # test
        >>> start = time.time()
        >>> multi_criterion.start_timing()
        >>> while multi_criterion.keep_running():
        ...     multi_criterion.iteration_done()
        >>> stop = time.time()
        >>> time_passed = stop - start
        >>> time_passed < 4
        True


    MaxIterationsTerminationCriterion stops the iterating:

    .. doctest::

        >>> from lclpy.termination.max_seconds_termination_criterion \\
        ...     import MaxSecondsTerminationCriterion
        >>> from lclpy.termination.max_iterations_termination_criterion \\
        ...     import MaxIterationsTerminationCriterion
        >>> from lclpy.termination.no_improvement_termination_criterion \\
        ...     import NoImprovementTerminationCriterion
        >>> from lclpy.termination.multi_criterion import MultiCriterion
        ... # init list
        >>> criteria = []
        >>> criteria.append(MaxSecondsTerminationCriterion(3))
        >>> criteria.append(MaxIterationsTerminationCriterion(10))
        >>> criteria.append(NoImprovementTerminationCriterion(3))
        ... # init MultiCriterion
        >>> multi_criterion = MultiCriterion(criteria)
        ... # test
        >>> iterations = 0
        >>> values = [20, 19, 18, 17, 16, 15, 14, 13, 12, 11, 10, 9]
        >>> multi_criterion.start_timing()
        >>> while multi_criterion.keep_running():
        ...     multi_criterion.check_new_value(values[iterations])
        ...     iterations += 1
        ...     multi_criterion.iteration_done()
        >>> iterations
        10

    NoImprovementTerminationCriterion stops the iterating:

    .. doctest::

        >>> from lclpy.termination.max_seconds_termination_criterion \\
        ...     import MaxSecondsTerminationCriterion
        >>> from lclpy.termination.max_iterations_termination_criterion \\
        ...     import MaxIterationsTerminationCriterion
        >>> from lclpy.termination.no_improvement_termination_criterion \\
        ...     import NoImprovementTerminationCriterion
        >>> from lclpy.termination.multi_criterion import MultiCriterion
        ... # init list
        >>> criteria = []
        >>> criteria.append(MaxSecondsTerminationCriterion(3))
        >>> criteria.append(MaxIterationsTerminationCriterion(10))
        >>> criteria.append(NoImprovementTerminationCriterion(3))
        ... # init MultiCriterion
        >>> multi_criterion = MultiCriterion(criteria)
        ... # test 1
        >>> iterations = 0
        >>> values = [9, 8, 7, 9, 9, 9, 9, 9, 9, 9, 9, 9]
        >>> multi_criterion.start_timing()
        >>> while multi_criterion.keep_running():
        ...     multi_criterion.check_new_value(values[iterations])
        ...     iterations += 1
        ...     multi_criterion.iteration_done()
        >>> iterations
        6


    """

    def __init__(self, criteria):
        super().__init__()
        self.criteria = criteria

    def keep_running(self):
        """function to determine if the algorithm needs to continue running.

        Returns
        -------
        bool
            The function returns true if the algorithm has to continue
            running, if the function returns false the algorithm needs to
            stop running. If one or more of the composing termination
            criterions returns False if its keep_running method is called, this
            method will return False. Else the method will return True.

        """

        for criterion in self.criteria:
            if criterion.keep_running() is False:
                return False
        return True

    def iteration_done(self):
        """function to be called after every iteration."""

        for criterion in self.criteria:
            criterion.iteration_done()

    def check_new_value(self, value):
        """Checks a value.

        Parameters
        ----------
        value : int or float
            A value from the evaluation function.

        """

        for criterion in self.criteria:
            criterion.check_new_value(value)

    def start_timing(self):
        """Starts an internal timer if needed."""

        for criterion in self.criteria:
            criterion.start_timing()

    def check_variable(self, variable):
        """Checks a variable specific to an implementation.

        Does not need to be used or implemented

        Parameters
        ----------
        variable
            The value of a certain value of a specific algorithm.

        """

        for criterion in self.criteria:
            criterion.check_variable(variable)

    def reset(self):
        """Resets the object back to it's state after init.

        Examples
        --------
         MaxIterationsTerminationCriterion stops the iterating:

        .. doctest::

            >>> from lclpy.termination.max_seconds_termination_criterion \\
            ...     import MaxSecondsTerminationCriterion
            >>> from lclpy.termination.max_iterations_termination_criterion \\
            ...     import MaxIterationsTerminationCriterion
            >>> from lclpy.termination.no_improvement_termination_criterion \\
            ...     import NoImprovementTerminationCriterion
            >>> from lclpy.termination.multi_criterion \\
            ...     import MultiCriterion
            ... # init list
            >>> criteria = []
            >>> criteria.append(MaxSecondsTerminationCriterion(3))
            >>> criteria.append(MaxIterationsTerminationCriterion(10))
            >>> criteria.append(NoImprovementTerminationCriterion(3))
            ... # init MultiCriterion
            >>> multi_criterion = MultiCriterion(criteria)
            ... # run 1
            >>> iterations = 0
            >>> values = [20, 19, 18, 17, 16, 15, 14, 13, 12, 11, 10, 9]
            >>> multi_criterion.start_timing()
            >>> while multi_criterion.keep_running():
            ...     multi_criterion.check_new_value(values[iterations])
            ...     iterations += 1
            ...     multi_criterion.iteration_done()
            >>> iterations
            10
            >>> # reset
            >>> multi_criterion.reset()
            ... # run 2
            >>> iterations = 0
            >>> values = [20, 19, 18, 17, 16, 15, 14, 13, 12, 11, 10, 9]
            >>> multi_criterion.start_timing()
            >>> while multi_criterion.keep_running():
            ...     multi_criterion.check_new_value(values[iterations])
            ...     iterations += 1
            ...     multi_criterion.iteration_done()
            >>> iterations
            10

        """

        for criterion in self.criteria:
            criterion.reset()



from lclpy.localsearch.move.array_swap import ArraySwap
import random


class TspArraySwap(ArraySwap):
    """Implements a swap move for 1 dimensional numpy arrays for tsp problems.

    In TSP problems swapping the first location with another is pointless. This
    class does not generate swaps with the first location from the
    neighbourhood. In all other respects this class is identical to it's base
    class Arrayswap.

    The move function performs and generates moves that swap 2 values in a
    one-dimensional array. Note that a move is represented as a tuple of int.
    The move (x, y) represents the swap of the values from the indices x and y
    of the array.

    Parameters
    ----------
    size : int
        The size of the numpy array that will be altered.
    Attributes
    ----------
    _size : int
        The size of the numpy array that is altered.

    Examples
    --------
    Get all possible moves, you should NEVER do this. You should evaluate only
    one move at a time. This example is simply to show the behaviour of
    get_moves and how to perform and undo a move:

    ..doctest::

        >>> import numpy
        >>> from lclpy.localsearch.move.tsp_array_swap import TspArraySwap
        ... # init array, a move will be performed on this array
        >>> array = numpy.array([0, 1, 2, 3, 4])
        ... # init
        >>> swap = TspArraySwap(len(array))
        ... # get all possible moves in all_moves
        >>> all_moves = []
        >>> for move in swap.get_moves():
        ...     all_moves.append(move)
        >>> all_moves
        [(1, 2), (1, 3), (1, 4), (2, 3), (2, 4), (3, 4)]
        >>> # picking an arbitrary move
        >>> # Never pick a move like this yourself. It only is done here for
        >>> # the sake of showing you a clear example.
        >>> a_move = all_moves[1]
        >>> a_move
        (1, 3)
        >>> # performing the move on the array
        >>> swap.move(array, a_move)
        >>> array
        array([0, 3, 2, 1, 4])
        >>> # undoing the move on the array
        >>> swap.undo_move(array, a_move) # undoes the move
        >>> array
        array([0, 1, 2, 3, 4])

    An example of generating some random moves with get_random_move:

    .. doctest::

        >>> import random
        >>> from lclpy.localsearch.move.tsp_array_swap import TspArraySwap
        ... # set seed random
        ... # not needed, is only used here to always get the same moves.
        >>> random.seed(0)
        ... # init
        >>> swap = TspArraySwap(10)
        ... # tests
        >>> swap.get_random_move()
        (1, 7)
        >>> swap.get_random_move()
        (5, 9)
        >>> swap.get_random_move()
        (7, 8)
        >>> swap.get_random_move()
        (5, 8)

    """

    def __init__(self, size):
        super().__init__(size)

    def get_moves(self):
        """This is a generator used to return all valid moves.

        Note that the swaps with the first position aren't included. When
        solving TSP problems, the start position doesn't matter.

        Yields
        ------
        tuple of int
            The next valid move.

        """

        for i in range(1, self._size):
            for j in range(i + 1, self._size):
                yield (i, j)

    def get_random_move(self):
        """This method is used to generate one random move.

        Note that the swaps with the first position aren't included. When
        solving TSP problems, the start position doesn't matter.

        Returns
        -------
        tuple of int
            A random valid move.
        """

        # It's possible to simply generate an i and then generate a bigger j,
        # but this wouldn't give us a proper distribution. Moves with a bigger
        # i would have a higher chance to be chosen than those with a smaller
        # i.

        # generate random numbers
        i = random.randrange(1, self._size)
        j = random.randrange(1, self._size)

        # ensure the number are different
        while i == j:
            j = random.randrange(1, self._size)

        # Puts the smallest number first, not needed, but ensures that every
        # move will have only 1 representation that will be used.
        if j < i:
            i, j = j, i

        return (i, j)



"""This package contains all the functions and classes needed for
delta-evaluation.

The easiest and fastest way to implement delta evaluation for new classes and
moves is to implement delta_evaluate in the evaluation function for a specific
evaluation-move class pair. The limitations of this approach are
obvious.

Alternatively, one can define it's own function that acts like delta_eval_func.
Obviously, one would need to use this function in the __init__ of your
evaluation classes instead of delta_eval_func. This means that the evaluation
functions defined by the library would need to be subclassed and given a
different __init__ method to be compatible with your function. The best way to
make your function use the already existing functionality would be like this:

.. code-block:: python

    def my_delta_eval_func(problem_eval_func, move_func):

        try:
            # try using existing functionality
            return delta_eval_func(problem_eval_func, move_func)

        except NotImplementedError:
            # a NotImplementedError was raised, so we know that delta_eval_func
            # couldn't find an implementation for a specific pair.

            # return of delta evaluation class for added classes you have
            # implemented
            return my_other_func(problem_eval_func, move_func)

Note that it might be possible to reuse classes, functions and code from
modules delta_qap, delta_tsp and delta_multi_neighbourhood.

"""



from abc import ABC, abstractmethod


class AbstractLocalSearch(ABC):
    """Template for the implementation of local search algorithms.

    This class should only contain the logic needed to run the local search.
    I/O and plotting should be handled by an AbstractRunner.

    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def run():
        """start running the local search

        Returns
        -------
        collections.namedtuple
            Contains the best found case and the evaluation value for said
            case.

        """

        pass

    def reset(self):
        """Resets the object back to it's state after init.

        Raises
        ------
        NotImplementedError
            Will always be raised.

        """

        raise NotImplementedError



from random import seed


def benchmark(problems, algorithms, stop_criterion, runs=10, seeds=None):
    """A function to perform multiple algorithms on multiple soltions.

    Note that the problems, algorithms and the stop criterion all need to have
    the method reset method properly implemented for this function to work
    properly. The logging of algorithms is determined by the logging parameter
    that was given to them at their construction. If it was True (default), the
    algorithm will log it's improvements and worse passed solutions. If logging
    is False, no logging will be shown for that particular algorithm.

    Parameters
    ----------
    problems : iterable object
        Contains all the problems.
    algorithms : iterable object
        Contains all the algorithms. Note that the algorithms can be
        initialised with None as solution and None as termination_criterion.
    stop_criterion : AbstractTerminationCriterion
        The termination criterion that will be used for all combinations of
        algorithms and problems.
    runs : int, optional
        The amount of runs that will be performed for a single
        algorithm-problem pair.
    seeds : list of int or tuple of int
        The seeds that will be used in the runs. Note that the length of the
        tuple or array needs to be equal to the amount of runs. If no seeds are
        given the seeds will be the number of the run.

    Returns
    -------
    list of list of list of namedtuple
        A 3-dimensional list of namedtuple. These namedtuples are the results
        of the algorithms. The first indice represents an algorithm, the second
        a problem, the third a run of the algorithm-problem pair. The indices
        that should be used are the same as in algorithms and solutions
        respectively for the first 2 indices. The third indice is used to
        choose between the runs. The possible indices for runs are always in
        the interval [0, #runs-1].

    """

    # set seeds if needed
    if seeds is None:
        seeds = range(runs)

    results = []

    algorithm_number = 0
    problem_number = 0
    seed_number = 0

    print('____Benchmark started___')
    # run everything
    for algorithm in algorithms:

        results_single_algorithm = []

        print('|---  Starting runs for algorithm ' + str(algorithm_number))

        for problem in problems:

            # setting problems and termination criterions
            algorithm._problem = problem
            algorithm._termination_criterion = stop_criterion

            different_seed_results = []

            print('--|---  Starting runs for problem ' + str(problem_number))

            for i in seeds:
                print('----|---  Starting run for seed ' + str(i))
                seed(i)
                algorithm.reset()
                different_seed_results.append(algorithm.run())
                print('----|--- Completed run for seed ' + str(i))
                seed_number += 1

            results_single_algorithm.append(different_seed_results)

            print('--|--- Completed runs for problem ' + str(problem_number))
            problem_number += 1

        results.append(results_single_algorithm)
        print('|--- Completed runs for algorithm ' + str(algorithm_number))
        algorithm_number += 1

    print('____Benchmark ended___')

    return results



from time import time
from lclpy.termination.abstract_termination_criterion \
    import AbstractTerminationCriterion


class MaxSecondsTerminationCriterion(AbstractTerminationCriterion):
    """Termination criterion to terminate after a set amount of seconds.

    Note that this terminationcriterion isn't exact. It will only terminate
    the algorithm after iterating longer than the set time AND if an iteration
    is finished. The extra time that the algorithm will run depends on the
    duration of the last iteration.

    Parameters
    ----------
    max_seconds : int or float, optional
        The maximal amount of seconds passed. The default is 60 seconds.

    Attributes
    ----------
    _max_seconds : int or float
        The maximal amount of seconds passed.
    _seconds : float
        The amount of seconds passed since the start of the iterations
    _start : float
        The moment when the time starts to be measured in seconds measured
        from the epoch.

    Examples
    --------
    Running for 60 seconds (default):

    .. doctest::

        >>> import time
        >>> from lclpy.termination.max_seconds_termination_criterion \\
        ...     import MaxSecondsTerminationCriterion
        ... # init
        >>> test = MaxSecondsTerminationCriterion()
        ... # start and stop will be used to measure the time passed.
        >>> start = time.time()
        ... # start the timing of the termination criterion
        >>> test.start_timing()
        ... # loop
        >>> while test.keep_running():
        ...     pass # code to execute
        ...     test.iteration_done()
        ... # calculate and check the time passed
        >>> stop = time.time()
        >>> time_passed = stop - start
        >>> time_passed < 61
        True

    Running for 3 seconds:

    .. doctest::

        >>> import time
        >>> from lclpy.termination.max_seconds_termination_criterion \\
        ...     import MaxSecondsTerminationCriterion
        ... # init
        >>> test = MaxSecondsTerminationCriterion(3)
        ... # start and stop will be used to measure the time passed.
        >>> start = time.time()
        ... # start the timing of the termination criterion
        >>> test.start_timing()
        ... # loop
        >>> while test.keep_running():
        ...     pass # code to execute
        ...     test.iteration_done()
        >>> stop = time.time()
        >>> time_passed = stop - start
        >>> time_passed < 4
        True

    """

    def __init__(self, max_seconds=60):
        super().__init__()
        self._start = 0
        self._max_seconds = max_seconds
        self._seconds = 0

    def keep_running(self):
        """function to determine if an algorithm needs to continue running

        Returns
        -------
        bool
            The function returns true if the amount of time passed is smaller
            than max_seconds, if the function returns false the amount of
            time passed is bigger than max_seconds

        """

        return self._seconds < self._max_seconds

    def start_timing(self):
        """function to be called before the iterations

        Sets _start to the current time in seconds from the epoch.

        """

        self._start = time()

    def iteration_done(self):
        """function to be called after every iteration

        Sets _seconds to be the difference between the current time and the
        start time.

        """

        self._seconds = time() - self._start

    def reset(self):
        """Resets the object back to it's state after init.

        Examples
        --------
        Running for 3 seconds:

        .. doctest::

            >>> import time
            >>> from lclpy.termination.max_seconds_termination_criterion \\
            ...     import MaxSecondsTerminationCriterion
            ... # init
            >>> test = MaxSecondsTerminationCriterion(3)
            ... # start and stop will be used to measure the time passed.
            ... # run 1
            >>> start = time.time()
            ... # start the timing of the termination criterion
            >>> test.start_timing()
            ... # loop
            >>> while test.keep_running():
            ...     pass # code to execute
            ...     test.iteration_done()
            >>> stop = time.time()
            >>> time_passed = stop - start
            >>> time_passed < 4
            True
            >>> # reset
            >>> test.reset()
            ... # run 2
            >>> start = time.time()
            ... # start the timing of the termination criterion
            >>> test.start_timing()
            ... # loop
            >>> while test.keep_running():
            ...     pass # code to execute
            ...     test.iteration_done()
            >>> stop = time.time()
            >>> time_passed = stop - start
            >>> time_passed < 4
            True



        """

        self._start = 0
        self._seconds = 0



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="lclpy",
    version="1.0.0",
    author="Daan Thijs",
    author_email="daan.thijs@hotmail.com",
    description="A localsearch library.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nobody1570/lclpy",
    packages=setuptools.find_packages(),
    install_requires=[
        'matplotlib>=3.0.2',
        'numpy>=1.15.4',
    ],
    python_requires=">=3.7",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
    ],
)



from lclpy.termination.abstract_termination_criterion \
    import AbstractTerminationCriterion
from lclpy.aidfunc.is_improvement_func import bigger, smaller


class NoImprovementTerminationCriterion(AbstractTerminationCriterion):
    """Criterion to terminate after a set amount of iterations without improvement.

    Parameters
    ----------
    max_iterations : int, optional
        The maximal amount of iterations without improvement. The default is
        100 iterations.
    minimise : bool, optional
        If the goal is to minimise the evaluation function, this should be
        True. If the goal is to maximise the evlauation function, this should
        be False. The default is True.

    Attributes
    ----------
    _max_iterations : int
        The maximal amount of iterations without improvement.
    _iterations : int
        The amount of iterations with no improvement.
    _run : bool
        Will be True if the maximal amount of iterations without improvement
        hasn't been reached yet, will be false if this isn't the case.
    _old_best_value : int
        The last best value. Is initialised as minus infinite
        (improvement_is_bigger = True)
        or infinite (improvement_is_bigger = False)
    _function : function
        The function used to judge if a value is an improvement.

    Examples
    --------
    Default amount of iterations without improvement (100), always checking
    the new value. Smaller values are considered improvements(default). The
    dataset eval_values is generated. (1 improvement, 99 cases without
    improvement, 1 improvement, 110 cases without improvement) After this the
    iterations start:

    .. doctest::

        >>> import numpy
        >>> from lclpy.termination.no_improvement_termination_criterion \\
        ...     import NoImprovementTerminationCriterion
        ... # creation of an array that contains the values that will be given
        ... # to our termination criterion
        >>> eval_values = numpy.concatenate(
        ...     (numpy.array([1000]), numpy.random.randint(1000,2000, size=98),
        ...      numpy.array([20]), numpy.random.randint(20,40, size=110)))
        ... # index is used to get values from the array.
        ... # index is also used to count the amount of iterations
        >>> index=0
        ... # init
        >>> test = NoImprovementTerminationCriterion()
        ... # loop
        >>> while test.keep_running():
        ...     pass # other code to execute
        ...     # check the next value
        ...     test.check_new_value(eval_values[index])
        ...     pass # other code to execute
        ...     # counting iterations + increment index
        ...     index += 1
        ...     test.iteration_done()
        >>> index # == amount of iterations.
        200

    3 iterations without improvement, only checking improved values. Smaller
    values are considered improvements(default). The dataset eval_values is
    hardcoded:

    .. doctest::

        >>> import numpy
        >>> from lclpy.termination.no_improvement_termination_criterion \\
        ...     import NoImprovementTerminationCriterion
        ... # creation of an array that contains the values that will be given
        ... # to our termination criterion
        >>> eval_values = numpy.array([20, 19, 21, 22, 23, 3])
        ... # index is used to get values from the array.
        ... # index is also used to count the amount of iterations
        >>> index = 0
        ... # old value is the last encountered best value
        >>> old_value = -1
        ... # init
        >>> test = NoImprovementTerminationCriterion(3)
        ... # loop
        >>> while test.keep_running():
        ...     pass # other code to execute
        ...     # only check next value if it's better than the last best value
        ...     if eval_values[index] > old_value:
        ...         old_value = eval_values[index]
        ...         test.check_new_value(eval_values[index])
        ...     pass # other code to execute
        ...     # counting iterations + increment index
        ...     index += 1
        ...     test.iteration_done()
        >>> index # == amount of iterations.
        4

    Default amount of iterations without improvement (100). Bigger values are
    considered improvements. The dataset eval_values is generated.
    (1 improvement, 99 cases without improvement, 1 improvement, 110 cases
    without improvement) After this the iterations start:

    .. doctest::

        >>> import numpy
        >>> from lclpy.termination.no_improvement_termination_criterion \\
        ...     import NoImprovementTerminationCriterion
        ... # creation of an array that contains the values that will be given
        ... # to our termination criterion
        >>> eval_values = numpy.concatenate(
        ...     (numpy.array([2]), numpy.random.randint(3, size=98),
        ...      numpy.array([5]), numpy.random.randint(6, size=110)))
        ... # index is used to get values from the array.
        ... # index is also used to count the amount of iterations
        >>> index=0
        ... # init
        >>> test = NoImprovementTerminationCriterion(minimise=False)
        ... # loop
        >>> while test.keep_running():
        ...     pass # other code to execute
        ...     # check next value
        ...     test.check_new_value(eval_values[index])
        ...     pass # other code to execute
        ...     # counting iterations + increment index
        ...     index += 1
        ...     test.iteration_done()
        >>> index # == amount of iterations.
        200

    3 iterations without improvement. Bigger values are considered
    improvements. The dataset eval_values is hardcoded:

    .. doctest::

        >>> import numpy
        >>> from lclpy.termination.no_improvement_termination_criterion \\
        ...     import NoImprovementTerminationCriterion
        ... # creation of an array that contains the values that will be given
        ... # to our termination criterion
        >>> eval_values = numpy.array([0, 0, 2, 1, 3, 3, 3, 4, 2, 1, 0, 12])
        ... # index is used to get values from the array.
        ... # index is also used to count the amount of iterations
        >>> index = 0
        ... # init
        >>> test = NoImprovementTerminationCriterion(3, False)
        ... # loop
        >>> while test.keep_running():
        ...     pass # other code to execute
        ...     # check next value
        ...     test.check_new_value(eval_values[index])
        ...     pass # other code to execute
        ...     # counting iterations + increment index
        ...     index += 1
        ...     test.iteration_done()
        >>> index # == amount of iterations.
        11




    """

    def __init__(self, max_iterations=100, minimise=True):
        super().__init__()

        # init
        self._max_iterations = max_iterations
        self._iterations = 0
        self._run = True

        # choose intial _old_best_value value + pick judge function
        if minimise:
            self._function = smaller
            self._old_best_value = float("inf")
        else:
            self._function = bigger
            self._old_best_value = float("-inf")

    def keep_running(self):
        """Function to determine if the algorithm needs to continue running.

        Returns
        -------
        bool
            The function returns true if the amount of iterations is smaller
            than _max_iterations, if the function returns false the amount of
            iterations is bigger than max_iterations

        """
        return self._run

    def iteration_done(self):
        """Function to be called after every iteration.

        Increments _iterations by 1.

        """

        self._iterations += 1

        if self._iterations == self._max_iterations:
            self._run = False

    def check_first_value(self, value):
        """Function that should be called once before the main loop.

        Parameters
        ----------
        value : int or float
            Is the evaluation value of the initial solution.

        """

        self._old_best_value = value

    def check_new_value(self, value):
        """Function to be called after every improvement of the evaluation function.

        It's also possible to call this function every time when the
        evaluation value is calculated without ill effects.

        Parameters
        ----------
        value : int or float
            Is the best evaluation value found for a solution or the new
            evaluation value of a solution. It does not matter which one is
            used.

        """
        if self._function(self._old_best_value, value):
            self._iterations = -1
            self._old_best_value = value

    def reset(self):
        """Resets the object back to it's state after init.

        Examples
        --------

        3 iterations without improvement. Bigger values are considered
        improvements. The dataset eval_values is hardcoded. After this, the
        criterion is resetted and the loop is repeated:

        .. doctest::

            >>> import numpy
            >>> from lclpy.termination.no_improvement_termination_criterion \\
            ...     import NoImprovementTerminationCriterion
            ... # creation of an array that contains the values that will be
            ... #  given to our termination criterion
            >>> eval_values = numpy.array(
            ...     [0, 0, 2, 1, 3, 3, 3, 4, 2, 1, 0, 12])
            ... # init
            >>> test = NoImprovementTerminationCriterion(3, False)
            ... # run 1
            ... #
            ... # index is used to get values from the array.
            ... # index is also used to count the amount of iterations
            >>> index = 0
            ... # loop
            >>> while test.keep_running():
            ...     pass # other code to execute
            ...     # check next value
            ...     test.check_new_value(eval_values[index])
            ...     pass # other code to execute
            ...     # counting iterations + increment index
            ...     index += 1
            ...     test.iteration_done()
            >>> index # == amount of iterations.
            11
            >>> # reset
            >>> test.reset()
            ... # run 2
            >>> index = 0
            ... # loop
            >>> while test.keep_running():
            ...     pass # other code to execute
            ...     # check next value
            ...     test.check_new_value(eval_values[index])
            ...     pass # other code to execute
            ...     # counting iterations + increment index
            ...     index += 1
            ...     test.iteration_done()
            >>> index # == amount of iterations.
            11

        """

        self._run = True
        self._iterations = 0

        # restore old_best value
        if self._function is smaller:
            # if minimising
            self._old_best_value = float("inf")
        else:
            # if maximising
            self._old_best_value = float("-inf")



import random
from lclpy.localsearch.move.abstract_move \
    import AbstractMove


class ArraySwap(AbstractMove):
    """Implements a swap move function for 1 dimensional numpy arrays.

    The move function performs and generates moves that swap 2 values in a
    one-dimensional array. Note that a move is represented as a tuple of int.
    The move (x, y) represents the swap of the values from the indices x and y
    of the array.

    Parameters
    ----------
    size : int
        The size of the numpy array that will be altered.

    Attributes
    ----------
    _size : int
        The size of the numpy array that is altered.

    Examples
    --------
    Get all possible moves, you should NEVER do this. You should evaluate only
    one move at a time. This example is simply to show the behaviour of
    get_moves and how to perform and undo a move:

    .. doctest::

        >>> import numpy
        >>> from lclpy.localsearch.move.array_swap import ArraySwap
        ... # init array, a move will be performed on this array
        >>> array = numpy.array([0, 1, 2, 3, 4])
        ... # init
        >>> swap = ArraySwap(len(array))
        ... # get all possible moves in all_moves
        >>> all_moves = []
        >>> for move in swap.get_moves():
        ...     all_moves.append(move)
        >>> all_moves
        [(0, 1), (0, 2), (0, 3), (0, 4), (1, 2), (1, 3), (1, 4), (2, 3), (2, 4), (3, 4)]
        >>> # picking an arbitrary move
        >>> # Never pick a move like this yourself. It only is done here for
        >>> # the sake of showing you a clear example.
        >>> a_move = all_moves[5]
        >>> a_move
        (1, 3)
        >>> # performing the move on the array
        >>> swap.move(array, a_move)
        >>> array
        array([0, 3, 2, 1, 4])
        >>> # undoing the move on the array
        >>> swap.undo_move(array, a_move)
        >>> array
        array([0, 1, 2, 3, 4])

    An example of generating some random moves with get_random_move:

    .. doctest::

        >>> import random
        >>> from lclpy.localsearch.move.array_swap import ArraySwap
        ... # set seed random
        ... # not needed, is only used here to always get the same moves.
        >>> random.seed(0)
        ... # init
        >>> swap = ArraySwap(10)
        ... # tests
        >>> swap.get_random_move()
        (0, 6)
        >>> swap.get_random_move()
        (4, 8)
        >>> swap.get_random_move()
        (6, 7)
        >>> swap.get_random_move()
        (4, 7)

    """

    def __init__(self, size):
        super().__init__()

        self._size = size

    def get_move_type(self):
        """Returns the move type.

        Returns
        -------
        str
            The move type.

        """

        return 'array_swap'

    def move(self, array, move):
        """Performs the move asked.

        Parameters
        ----------
        array : numpy.ndarray
            The array where items will be swapped.
        move : tuple of int
            Represents 1 unique move. Valid moves can be retrieved by using
            get_random_move and get_move.

        """

        (index_1, index_2) = move

        array[index_1], array[index_2] = array[index_2], array[index_1]

    def undo_move(self, array, move):
        """Undoes the move asked.

        Parameters
        ----------
        array : numpy.ndarray
            The array where items will be swapped.
        move : tuple of int
            Represents 1 unique move. Valid moves can be retrieved by using
            get_random_move and get_move.

        """

        self.move(array, move)

    def get_moves(self):
        """Iterate over all valid moves.

        Yields
        ------
        tuple of int
            The next valid move.

        """

        for i in range(self._size):
            for j in range(i + 1, self._size):
                yield (i, j)

    def get_random_move(self):
        """This method is used to generate one random move.

        Returns
        -------
        tuple of int
            A random valid move.
        """

        # It's possible to simply generate an i and then generate a bigger j,
        # but this wouldn't give us a proper distribution. Moves with a bigger
        # i would have a higher chance to be chosen than those with a smaller
        # i.

        # generate random numbers
        i = random.randrange(self._size)
        j = random.randrange(self._size)

        # ensure the number are different
        while i == j:
            j = random.randrange(self._size)

        # Puts the smallest number first, not needed, but ensures that every
        # move will have only 1 representation that will be used.
        if j < i:
            i, j = j, i

        return (i, j)



from lclpy.termination.multi_criterion import MultiCriterion


class MultiAndCriterion(MultiCriterion):
    """Class to use multiple terminationcriteria at once.

    Parameters
    ----------
    criteria : list or tuple of AbstractTerminationCriterion
        An iterable object containing the intialised termination criterions one
        wishes to use.

    Attributes
    ----------
    criteria : list or tuple of AbstractTerminationCriterion
        An iterable object containing the intialised termination criterions one
        wishes to use.

    Examples
    --------
    3 termination criteria are used, three tests are done to ensure that all
    three criterions are capable of stopping the iterating correctly.

    MaxSecondsTerminationCriterion stops the iterating:

     .. doctest::

        >>> import time
        >>> from lclpy.termination.max_seconds_termination_criterion \\
        ...     import MaxSecondsTerminationCriterion
        >>> from lclpy.termination.max_iterations_termination_criterion \\
        ...     import MaxIterationsTerminationCriterion
        >>> from lclpy.termination.no_improvement_termination_criterion \\
        ...     import NoImprovementTerminationCriterion
        >>> from lclpy.termination.multi_and_criterion import MultiAndCriterion
        ... # init list
        >>> criteria = []
        >>> criteria.append(MaxSecondsTerminationCriterion(0))
        >>> criteria.append(MaxIterationsTerminationCriterion(0))
        >>> criteria.append(NoImprovementTerminationCriterion(3))
        ... # init MultiCriterion
        >>> multi_criterion = MultiAndCriterion(criteria)
        ... # test
        >>> start = time.time()
        >>> multi_criterion.start_timing()
        >>> while multi_criterion.keep_running():
        ...     multi_criterion.iteration_done()
        >>> stop = time.time()
        >>> time_passed = stop - start
        >>> time_passed < 4
        True


    MaxIterationsTerminationCriterion stops the iterating:

    .. doctest::

        >>> from lclpy.termination.max_seconds_termination_criterion \\
        ...     import MaxSecondsTerminationCriterion
        >>> from lclpy.termination.max_iterations_termination_criterion \\
        ...     import MaxIterationsTerminationCriterion
        >>> from lclpy.termination.no_improvement_termination_criterion \\
        ...     import NoImprovementTerminationCriterion
        >>> from lclpy.termination.multi_and_criterion import MultiAndCriterion
        ... # init list
        >>> criteria = []
        >>> criteria.append(MaxSecondsTerminationCriterion(0))
        >>> criteria.append(MaxIterationsTerminationCriterion(10))
        >>> criteria.append(NoImprovementTerminationCriterion(3))
        ... # init MultiCriterion
        >>> multi_criterion = MultiAndCriterion(criteria)
        ... # test
        >>> iterations = 0
        >>> values = [20, 19, 18, 20, 20, 20, 20, 13, 12, 11, 10, 9]
        >>> multi_criterion.start_timing()
        >>> while multi_criterion.keep_running():
        ...     multi_criterion.check_new_value(values[iterations])
        ...     iterations += 1
        ...     multi_criterion.iteration_done()
        >>> iterations
        10

    NoImprovementTerminationCriterion stops the iterating:

    .. doctest::

        >>> from lclpy.termination.max_seconds_termination_criterion \\
        ...     import MaxSecondsTerminationCriterion
        >>> from lclpy.termination.max_iterations_termination_criterion \\
        ...     import MaxIterationsTerminationCriterion
        >>> from lclpy.termination.no_improvement_termination_criterion \\
        ...     import NoImprovementTerminationCriterion
        >>> from lclpy.termination.multi_and_criterion import MultiAndCriterion
        ... # init list
        >>> criteria = []
        >>> criteria.append(MaxSecondsTerminationCriterion(0))
        >>> criteria.append(MaxIterationsTerminationCriterion(2))
        >>> criteria.append(NoImprovementTerminationCriterion(3))
        ... # init MultiCriterion
        >>> multi_criterion = MultiAndCriterion(criteria)
        ... # test 1
        >>> iterations = 0
        >>> values = [9, 8, 7, 9, 9, 9, 9, 9, 9, 9, 9, 9]
        >>> multi_criterion.start_timing()
        >>> while multi_criterion.keep_running():
        ...     multi_criterion.check_new_value(values[iterations])
        ...     iterations += 1
        ...     multi_criterion.iteration_done()
        >>> iterations
        6

        """

    def keep_running(self):
        """function to determine if the algorithm needs to continue running.

        Returns
        -------
        bool
            The function returns true if the algorithm has to continue
            running, if the function returns false the algorithm needs to
            stop running. If all the composing termination criterions return
            False if their keep_running method is called, this method will
            return False. Else the method will return True.

        """

        for criterion in self.criteria:
            if criterion.keep_running() is True:
                return True
        return False



import string
import numpy
import math
import collections
import itertools

PI = 3.141592
RRR = 6378.388


def _euclidian(from_x, from_y, to_x, to_y):
    """Calculates the euclidian distance between 2 points in 2D space.

    Parameters
    ----------
    from_x : int or float
        The x coordinate of the 1st point.
    from_y : int or float
        The y coordinate of the 1st point.
    to_x : int or float
        The x coordinate of the 2nd point.
    to_y : int or float
        The y coordinate of the 2nd point.

    Returns
    -------
    int
        The euclidian distance between the 2 points. Rounded to the nearest
        int.


    """

    return int(round(math.sqrt((to_x - from_x)**2 + (to_y - from_y)**2)))


def _euclidian_rounded_up(from_x, from_y, to_x, to_y):
    """Calculates the euclidian distance between 2 points in 2D space.

    Parameters
    ----------
    from_x : int or float
        The x coordinate of the 1st point.
    from_y : int or float
        The y coordinate of the 1st point.
    to_x : int or float
        The x coordinate of the 2nd point.
    to_y : int or float
        The y coordinate of the 2nd point.

    Returns
    -------
    int
        The euclidian distance between the 2 points, rounded up to the nearest
        int.


    """

    return int(math.ceil(math.sqrt((to_x - from_x)**2 + (to_y - from_y)**2)))


def _manhattan(from_x, from_y, to_x, to_y):
    """Calculates the manhattan distance between 2 points in 2D space.

    Parameters
    ----------
    from_x : int or float
        The x coordinate of the 1st point.
    from_y : int or float
        The y coordinate of the 1st point.
    to_x : int or float
        The x coordinate of the 2nd point.
    to_y : int or float
        The y coordinate of the 2nd point.

    Returns
    -------
    int
        The manhattan distance between the 2 points. Rounded to the
        nearest int.


    """

    return int(round(abs(to_x - from_x) + abs(to_y - from_y)))


def _degrees_to_radian(angle_degrees):
    """Converts radian to degrees.

    The numbers behind the point are minutes, not degrees. This is simply a
    convention of tsplib, don't worry about it.

    Parameters
    ----------
    angle_degrees : float
        An angle in degrees.

    Returns
    -------
    float
        The angle in radian.

    """

    degrees = int(angle_degrees)
    minutes = angle_degrees - degrees

    return PI * (degrees + 5.0 * minutes / 3.0) / 180.0


def _geo(from_x, from_y, to_x, to_y):
    """Calculates the manhattan distance between 2 points on the globe.

    Parameters
    ----------
    from_x : int or float
        Latitude of the 1st point.
    from_y : int or float
        Longitude of the 1st point.
    to_x : int or float
        Latitude of the 2nd point.
    to_y : int or float
        Longitude of the 2nd point.

    Returns
    -------
    int
        The distance between the 2 points in km, rounded to the nearest int.


    """

    # convert degrees to radian

    from_x_radian = _degrees_to_radian(from_x)
    to_x_radian = _degrees_to_radian(to_x)

    from_y_radian = _degrees_to_radian(from_y)
    to_y_radian = _degrees_to_radian(to_y)

    # calculate distance

    q1 = math.cos(to_y_radian - from_y_radian)
    q2 = math.cos(to_x_radian - from_x_radian)
    q3 = math.cos(to_x_radian + from_x_radian)
    return int(
        RRR * math.acos(0.5 * ((1.0 + q1) * q2 - (1.0 - q1) * q3)) + 1.0)


def _att(from_x, from_y, to_x, to_y):
    """Calculates a special pseudo-euclidian distance between 2 points in 2D space.

    Parameters
    ----------
    from_x : int or float
        The x coordinate of the 1st point.
    from_y : int or float
        The y coordinate of the 1st point.
    to_x : int or float
        The x coordinate of the 2nd point.
    to_y : int or float
        The y coordinate of the 2nd point.

    Returns
    -------
    int
        The distance between the 2 points.


    """

    xd = to_x - from_x
    yd = to_y - from_y
    rij = math.sqrt((xd * xd + yd * yd) / 10.0)
    tij = int(round(rij))

    if tij < rij:
        tij += 1

    return tij


def _default_processing(data, dist_func, type=numpy.float_):
    """Creates a dict and calculates the distance matrix for a 2D tsp problem.

    Parameters
    ----------
    data : list of list
        The lists in the list always contain the name of a point,
        the x coordinate and the y coordinate in that order seperated by
        spaces.
    dist_func : function
        A function to calculate the distance between the points. This
        function's arguments must be the x and y coordinates of the first
        point, followed by the x and y coordinates of the second point.
    type : numpy.dtype, optional
        The data type used by the numpy array. The default is numpy.float_,
        which is the default datatype when creating a numpy.ndarray.

    Returns
    -------
    dist_matrix : numpy.ndarray
        The distance matrix for the problem.
    dictionary : {int : int}
        A dictionary that can convert a position the distance matrix to the
        name given in data.


    """

    # make dist_matrix
    size = len(data)

    # dist_matrix [from] [to]
    dist_matrix = numpy.full((size, size), numpy.inf, dtype=type)

    for i in range(size):

        # ditance between a point and itself is always zero.
        dist_matrix[i][i] = 0

        # get coordinates point
        i_dist_x = data[i][1]
        i_dist_y = data[i][2]

        for j in range(i + 1, size):
            # get coordinates all points
            j_dist_x = data[j][1]
            j_dist_y = data[j][2]

            # calculate distance
            dist_matrix[i][j] = dist_func(
                i_dist_x, i_dist_y, j_dist_x, j_dist_y)
            dist_matrix[j][i] = dist_func(
                i_dist_x, i_dist_y, j_dist_x, j_dist_y)

    return dist_matrix


def _upper_row_processing(data, dimension):
    """Creates a dict and initialises the distance matrix for a 2D tsp problem.

    Parameters
    ----------
    data : list of list
        The lists in the list contain the upper row of the distance matrix
    dimension : int
        The dimension of the distance matrix.
    Returns
    -------
    dist_matrix : numpy.ndarray
        The distance matrix for the problem.

    """

    # make distance matrix
    # all values of matrix are initialised as 0
    dist_matrix = numpy.full((dimension, dimension), 0, dtype=numpy.int_)

    # create iterator to iterate over all ints in list
    iterator = itertools.chain.from_iterable(data)

    # fill the distance matrix
    for i in range(dimension):
        for j in range(i, dimension):

            # skip first diagonal --> will always be 0
            if i != j:
                value = int(next(iterator))

                # if values of the first diagonal are included, skip them.
                if value == 0:
                    value = int(next(iterator))

                dist_matrix[i][j] = int(value)
                dist_matrix[j][i] = int(value)

    return dist_matrix


def _lower_row_processing(data, dimension):
    """Creates a dict and initialises the distance matrix for a 2D tsp problem.

    Parameters
    ----------
    data : list of string
        The strings in the list contain the lower row of the distance matrix
    dimension : int
        The dimension of the distance matrix.
    Returns
    -------
    dist_matrix : numpy.ndarray
        The distance matrix for the problem.

    """

    # make distance matrix
    # all values of matrix are initialised as 0
    dist_matrix = numpy.full((dimension, dimension), 0, dtype=numpy.int_)

    # create iterator to iterate over all ints in list
    iterator = itertools.chain.from_iterable(data)

    # fill the distance matrix
    for i in range(dimension):
        for j in range(i + 1):

            # skip first diagonal --> will always be 0
            if i != j:
                value = int(next(iterator))

                # if values of the first diagonal are included, skip them.
                if value == 0:
                    value = int(next(iterator))

                dist_matrix[i][j] = int(value)
                dist_matrix[j][i] = int(value)

    return dist_matrix


def _matrix_processing(data, dimension):
    """Creates a dict and initialises the distance matrix for a 2D tsp problem.

    Parameters
    ----------
    data : list of string
        The strings in the list contain the full distance matrix
    dimension : int
        The dimension of the distance matrix.
    Returns
    -------
    dist_matrix : numpy.ndarray
        The distance matrix for the problem.

    """

    # make distance matrix
    dist_matrix = numpy.full((dimension, dimension), 0, dtype=numpy.int_)

    # create iterator to iterate over all ints in list
    iterator = itertools.chain.from_iterable(data)

    # fill matrix
    for i in range(dimension):
        for j in range(dimension):

            value = int(next(iterator))
            dist_matrix[i][j] = int(value)
            dist_matrix[j][i] = int(value)

    return dist_matrix


def read_tsplib(filename):
    """Converts the tsplib file format to useable data structures.

    Currently this function only works for TSP problems. The crystallography
    problems can't be parsed with this function.

    Parameters
    ----------
    filename : str
        absolute or relative path to the file that contains the data.

    Returns
    -------
    distance_matrix : numpy.ndarray
        The distance matrix for the problem.
    dictionary : {int : int}
        A dictionary that can convert a position the distance matrix to the
        name given in the data. This dict can be useful when generating output.
    metadata : list of str
        Contains the metadata of the problem. The last entry will always be
        EOF.

    Examples
    --------
    Read "testfile.tsp" from the map "data" in the current working directory:

    .. code-block:: python

        from lclpy.io.tsplib import read_tsplib
        read_tsplib('data/testfile.tsp')

    """

    # init
    metadata = []
    data = []

    # read data
    with open(filename) as f:
        for line in f:
            if line[:1] in string.ascii_uppercase:
                metadata.append(line)
            else:
                splitted = line.split()
                numbers = []
                for i in splitted:
                    numbers.append(float(i))
                data.append(numbers)

    # check problem type and data type
    type_metadata = [s for s in metadata if 'TYPE' in s]

    # choose problem type

    dtype = None
    dist_func = None
    if 'TSP' in type_metadata[0]:
        solve = _default_processing

        if 'EUC_2D' in type_metadata[1]:
            dtype = numpy.int_
            dist_func = _euclidian
        elif 'MAN_2D' in type_metadata[1]:
            dtype = numpy.int_
            dist_func = _manhattan
        elif 'CEIL_2D' in type_metadata[1]:
            dtype = numpy.int_
            dist_func = _euclidian_rounded_up
        elif 'GEO' in type_metadata[1]:
            dtype = numpy.int_
            dist_func = _geo
        elif 'ATT' in type_metadata[1]:
            dtype = numpy.int_
            dist_func = _att

        elif any('EXPLICIT' in s for s in metadata):

            if any('UPPER_ROW' in s for s in metadata) \
                    or any('UPPER_DIAG_ROW' in s for s in metadata):
                dtype = numpy.int_
                solve = _upper_row_processing
                dimension_metadata = [s for s in metadata if 'DIMENSION' in s]
                dimension = int(dimension_metadata[0].split()[-1])

            elif any('LOWER_ROW' in s for s in metadata) \
                    or any('LOWER_DIAG_ROW' in s for s in metadata):
                dtype = numpy.int_
                solve = _lower_row_processing
                dimension_metadata = [s for s in metadata if 'DIMENSION' in s]
                dimension = int(dimension_metadata[0].split()[-1])

            elif any('FULL_MATRIX' in s for s in metadata):
                dtype = numpy.int_
                solve = _matrix_processing
                dimension_metadata = [s for s in metadata if 'DIMENSION' in s]
                dimension = int(dimension_metadata[0].split()[-1])
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError

    if dist_func is None:
        dist_matrix = solve(data, dimension)
    elif dtype is None:
        dist_matrix = solve(data, dist_func)
    else:
        dist_matrix = solve(data, dist_func, dtype)

    # make dictionary, is the same in all cases
    # tsplib files start counting from 1, not from 0
    dictionary = {}

    for i in range(dist_matrix.shape[0]):
        dictionary[i] = i + 1

    # return results
    TsplibData = collections.namedtuple(
        'TsplibData', ['distance_matrix', 'dictionary', 'metadata'])
    return TsplibData(dist_matrix, dictionary, metadata)



from abc import ABC, abstractmethod


class AbstractTerminationCriterion(ABC):
    """Template to create termination criterions."""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def keep_running(self):
        """Function to determine if the algorithm needs to continue running

        Returns
        -------
        bool
            The function returns true if the algorithm has to continue
            running, if the function returns false the algorithm needs to
            stop running.

        """

        pass

    def iteration_done(self):
        """Function to be called after every iteration.

        Does not need to be used or implemented.

        """

        pass

    def check_first_value(self, value):
        """Function that should be called once before the main loop.

        Parameters
        ----------
        value : int or float
            Is the evaluation value of the initial problem.

        """

        pass

    def check_new_value(self, value):
        """Checks a value.

        Does not need to be used or implemented.

        Parameters
        ----------
        value : int or float
            A value from the evaluation function.

        """

        pass

    def check_variable(self, variable):
        """Checks a variable specific to an implementation.

        Does not need to be used or implemented

        Parameters
        ----------
        variable
            The value of a certain value of a specific algorithm.

        """

        pass

    def start_timing(self):
        """Starts an internal timer if needed.

        Does not need to be used or implemented.

        """

        pass

    def reset(self):
        """Resets the object back to it's state after init."""

        raise NotImplementedError





def log_improvement(value):
    """function to log improvements to the command line.

    Parameters
    ----------
    value : int or float
            The value for the improvement

    """

    print("Improvement : " + str(value))


def log_passed_worse(value):
    """function to log the passing of worse solutions to the command line.

    Parameters
    ----------
    value : int or float
            The value for the improvement

    """

    print("Passed worse: " + str(value))



from lclpy.termination.abstract_termination_criterion \
    import AbstractTerminationCriterion


class MaxIterationsTerminationCriterion(AbstractTerminationCriterion):
    """Termination criterion to terminate after a set amount of iterations.

    Parameters
    ----------
    max_iterations : int, optional
        The maximal amount of iterations. The default is 1000 iterations.

    Attributes
    ----------
    _max_iterations : int
        The maximal amount of iterations.
    _iterations : int
        The amount of iterations.

    Examples
    --------
    Default amount of iterations:

    .. doctest::

        >>> from lclpy.termination.max_iterations_termination_criterion \\
        ...     import MaxIterationsTerminationCriterion
        ... # the variable iterations will be used to count the amount of
        ... # iterations.
        >>> iterations = 0
        ... # init
        >>> test = MaxIterationsTerminationCriterion()
        ... # test
        >>> while test.keep_running():
        ...     pass # other code to be executed
        ...     # counting amount of iterations
        ...     iterations += 1
        ...     test.iteration_done()
        >>> iterations
        1000

    50 iterations:

    .. doctest::

        >>> from lclpy.termination.max_iterations_termination_criterion \\
        ...     import MaxIterationsTerminationCriterion
        ... # the variable iterations will be used to count the amount of
        ... # iterations.
        >>> iterations = 0
        ... # init
        >>> test = MaxIterationsTerminationCriterion(50)
        ... # test
        >>> while test.keep_running():
        ...     pass # other code to be executed
        ...     # counting amount of iterations
        ...     iterations += 1
        ...     test.iteration_done()
        >>> iterations
        50

    """

    def __init__(self, max_iterations=1000):
        super().__init__()
        self._max_iterations = max_iterations
        self._iterations = 0

    def keep_running(self):
        """function to determine if the algorithm needs to continue running.

        Returns
        -------
        bool
            The function returns true if the amount of iteration is smaller
            than max_iterations, if the function returns false the amount of
            iterations is bigger than max_iterations

        """

        return self._iterations < self._max_iterations

    def iteration_done(self):
        """function to be called after every iteration.

        Increments _iterations by 1.

        """

        self._iterations += 1

    def reset(self):
        """Resets the object back to it's state after init.

        Examples
        --------
        Default amount of iterations:

        .. doctest::

            >>> from lclpy.termination.max_iterations_termination_criterion \\
            ...     import MaxIterationsTerminationCriterion
            ... # the variable iterations will be used to count the amount of
            ... # iterations.
            >>> iterations = 0
            ... # init
            >>> test = MaxIterationsTerminationCriterion()
            ... # run 1
            >>> while test.keep_running():
            ...     pass # other code to be executed
            ...     # counting amount of iterations
            ...     iterations += 1
            ...     test.iteration_done()
            >>> iterations
            1000
            >>> # reset
            >>> test.reset()
            ... # run 2
            >>> iterations = 0
            >>> while test.keep_running():
            ...     pass # other code to be executed
            ...     # counting amount of iterations
            ...     iterations += 1
            ...     test.iteration_done()
            >>> iterations
            1000

        """

        self._iterations = 0
